"""These class are here to help unitesting the database classes"""
import os
import typing
from typing import List

import pytest
from koalak.utils import get_prefixed_callables_of_object, temp_pathname

from .databases import ListDB


class HelperTest:
    def test(self):
        for test_method in get_prefixed_callables_of_object(self, "test_"):
            test_method()


class HelperTestListDatabase(HelperTest):
    def __init__(self, cls: typing.Type[ListDB], values: List = None):
        if values is None:
            values = [1, 2, 3, "c", True, [], {}]

        self.cls = cls
        self.values = values

        # Values to test with must be at least 3
        assert len(self.values) >= 3

        # Must be unique
        # convert to string so that we can transform our list into set
        #  because dict can not be hashable
        tmp_values = [str(e) for e in self.values]
        assert len(set(tmp_values)) == len(self.values)

        # The first three element must be sorted
        assert self.values[:3] == sorted(self.values[:3])

    def test_simple(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as path:
            l = self.cls(path)

            l.append(v1)

            assert len(l) == 1
            assert l[0] == v1
            assert list(l) == [v1]
            assert l.count(v1) == 1

            # Retest without append
            l = self.cls(path)
            assert len(l) == 1
            assert l[0] == v1
            assert list(l) == [v1]
            assert l.count(v1) == 1

            # Modify it
            l = self.cls(path)
            l[0] = v2

            l = self.cls(path)
            assert l[0] == v2

    def test__add_radd(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as path:
            l = self.cls(path, [v1], autosync=True)
            l2 = l + [v2]

            assert l2 == [v1, v2]
            assert type(l2) is list
            assert list(l) == [v1]

            l3 = [v2] + l
            assert l3 == [v2, v1]
            assert type(l3) is list
            assert list(l) == [v1]

    def test__del(self):

        v1, v2, v3 = self.values[:3]

        with temp_pathname() as path:

            l = self.cls(path, [v1, v2, v3], autosync=True)
            assert list(l) == [v1, v2, v3]

            del l[0]
            assert list(l) == [v2, v3]
            l = self.cls(path, autosync=True)
            assert list(l) == [v2, v3]

    def test__getitem(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as path:
            l = self.cls(path, [v1, v2, v3], autosync=True)
            assert list(l) == [v1, v2, v3]

            assert l[0] == v1
            assert l[-1] == v3

            assert l[:] == [v1, v2, v3]
            assert type(l[:]) is list
            assert type(l[:2]) is list  # and not DBList
            assert l[:2] == [v1, v2]

    def test_pop(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as path:
            l = self.cls(path, [v1, v2, v3], autosync=True)
            l.pop()
            assert l == [v1, v2]
            l = self.cls(path, autosync=True)
            assert l == [v1, v2]

            l.pop(0)
            assert l == [v2]
            l = self.cls(path, autosync=True)
            assert l == [v2]

    def test__mul_rmul(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as path:
            l = self.cls(path, [v1], autosync=True)
            l2 = l * 2

            assert l2 == [v1, v1]
            assert type(l2) is list
            assert list(l) == [v1]

            l3 = 2 * l
            assert l3 == [v1, v1]
            assert type(l3) is list
            assert list(l) == [v1]

    def test_count(self):
        v1 = self.values[0]
        v2 = self.values[1]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            l.append(v1)

            assert l.count(v1) == 1
            l = self.cls(dbpath, autosync=True)
            assert l.count(v1) == 1

            l.append(v2)
            l.append(v1)
            assert l.count(v1) == 2
            assert l.count(v2) == 1

            l = self.cls(dbpath, autosync=True)
            assert l.count(v1) == 2
            assert l.count(v2) == 1

    def test_index(self):
        v1 = self.values[0]
        v2 = self.values[1]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            l.append(v1)
            l.append(v2)

            assert l.index(v1) == 0
            assert l.index(v2) == 1

            l = self.cls(dbpath, autosync=True)
            assert l.index(v1) == 0
            assert l.index(v2) == 1

            # example with start/stop
            assert l.index(v2, 1) == 1
            assert l.index(v1, 0, 2) == 0

    def test_autosync(self):
        v1 = self.values[0]
        v2 = self.values[1]
        v3 = self.values[2]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            l.append(v1)

            # append
            l = self.cls(dbpath, autosync=True)
            assert l[0] == v1

            # remove
            l.remove(v1)
            l = self.cls(dbpath, autosync=True)
            assert len(l) == 0

            # test extend
            l.extend(self.values[:3])  # extend the 3 first elements
            l = self.cls(dbpath, autosync=True)
            assert len(l) == 3
            assert list(l) == self.values[:3]

            # test clear
            l.clear()
            l = self.cls(dbpath, autosync=True)
            assert list(l) == []

            # test insert
            l.clear()
            l.append(v2)
            l.insert(0, v1)
            assert list(l) == [v1, v2]
            l = self.cls(dbpath, autosync=True)
            assert list(l) == [v1, v2]

            # test pop
            l.clear()
            l.append(v1)
            l.append(v2)
            assert l.pop(0) == v1
            assert l.pop(0) == v2
            assert list(l) == []
            l = self.cls(dbpath, autosync=True)
            assert list(l) == []

            # test reverse
            l.clear()
            l.append(v1)
            l.append(v2)
            l.append(v3)
            l.reverse()
            assert list(l) == [v3, v2, v1]
            l = self.cls(dbpath, autosync=True)
            assert list(l) == [v3, v2, v1]

            # test sort
            l.clear()
            l.append(v2)
            l.append(v3)
            l.append(v1)
            l.sort()
            assert list(l) == [v1, v2, v3]
            l = self.cls(dbpath, autosync=True)
            assert list(l) == [v1, v2, v3]

    def test_contains(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            assert "a" not in l
            assert "b" not in l

            l.append("a")
            assert "a" in l
            l = self.cls(dbpath, autosync=True)
            assert "a" in l

            l.append("b")
            assert "a" in l
            assert "b" in l
            assert "c" not in l
            l = self.cls(dbpath, autosync=True)
            assert "a" in l
            assert "b" in l
            assert "c" not in l

            l.append("c")
            assert "a" in l
            assert "b" in l
            assert "c" in l
            l = self.cls(dbpath, autosync=True)
            assert "a" in l
            assert "b" in l
            assert "c" in l

            l.remove("a")
            assert "a" not in l
            l = self.cls(dbpath, autosync=True)
            assert "a" not in l

    def test_file_is_created(self):
        """Assert that the file is created in the constructor"""
        with temp_pathname() as dbpath:
            self.cls(dbpath, autosync=True)
            assert os.path.exists(dbpath)

    def test_initiated_with_existing_list(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, [v1, v2, v3], autosync=True)
            assert list(l) == [v1, v2, v3]
            l = self.cls(dbpath, autosync=True)
            assert list(l) == [v1, v2, v3]

            l.remove(v1)
            assert list(l) == [v2, v3]

            l = self.cls(dbpath, autosync=True)
            assert list(l) == [v2, v3]

    def test_conflict_args_obj_and_existing_db(self):
        v1, v2, v3 = self.values[:3]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)  # the file dbpath is created
            with pytest.raises(TypeError):
                # once the file is created we can't initiate it with an existing list
                l = [v1, v2, v3]
                self.cls(dbpath, l, autosync=True)

    def test_double_sync(self):
        """Test that double sync is possible"""
        v1, v2, v3 = self.values[:3]
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            l.append(v1)
            l.sync()
            l.sync()

    def test_nb_sync(self):
        v1, v2, v3 = self.values[:3]
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            nb_sync = l._nb_sync
            l.append(v1)  # sync called here
            l.sync()  # this sync is useless because already synced
            assert l._nb_sync == nb_sync + 1  # check that the previous sync did nothing
            l.extend([v1, v1, v1])
            assert (
                l._nb_sync == nb_sync + 2
            )  # check that extend call sync once and not 3 time

    def test_unique(self):
        v1, v2, v3 = self.values[:3]

        # test append
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            with pytest.raises(ValueError):
                l.append("a")

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l.append("a")

        # test insert
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            with pytest.raises(ValueError):
                l.insert(0, "a")

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l.insert(0, "a")

        # test extend
        with temp_pathname() as dbpath:
            # FIXME: if error occure on extend we don't validate first elem?
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            with pytest.raises(ValueError):
                l.extend(["b", "a"])

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l.extend(["b", "a"])

        with temp_pathname() as dbpath:
            # FIXME: if error occure on extend we don't validate first elem?
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            with pytest.raises(ValueError):
                l.extend(["b", "b"])

            l = self.cls(dbpath, autosync=True, unique=True)

        # test setitem
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            l.append("b")
            with pytest.raises(ValueError):
                l[1] = "a"

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l[1] = "a"

            assert list(l) == ["a", "b"]
            # it is allowed to modify the same value
            l[0] = "a"

    def test_unique_as_function(self):
        v1, v2, v3 = self.values[:3]

        # test append
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=lambda x: x.split("=")[0])

            l.append("a=1")
            l.append("b=2")
            l.append("c=3")
            assert list(l) == ["a=1", "b=2", "c=3"]
            l = self.cls(dbpath, autosync=True, unique=lambda x: x.split("=")[0])
            assert list(l) == ["a=1", "b=2", "c=3"]

            with pytest.raises(ValueError):
                l.append("a=5")


class HelperTestDictDB(HelperTest):
    def __init__(
        self, cls: typing.Type[ListDB], values: List = None, keys: List[str] = None
    ):
        if values is None:
            values = [1, 2, 3, "c", True, [], {}]

        if keys is None:
            keys = ["a", "b", "c"]
        self.cls = cls
        self.values = values
        self.keys = keys

        # Values to test with must be at least 3
        assert len(self.values) >= 3
        assert len(self.keys) >= 3
        # Must be unique
        # convert to string so that we can transform our list into set
        #  because dict can not be hashable
        tmp_values = [str(e) for e in self.values]
        assert len(set(tmp_values)) == len(self.values)

        tmp_keys = [str(e) for e in self.keys]
        assert len(set(tmp_keys)) == len(self.keys)

        # The first three element must be sorted
        assert self.values[:3] == sorted(self.values[:3])
        assert self.keys[:3] == sorted(self.keys[:3])

    def test_simple(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]

        with temp_pathname() as path:
            l = self.cls(path, autosync=True)

            l[k1] = v1

            assert len(l) == 1
            assert l[k1] == v1
            assert list(l) == [k1]
            assert dict(l) == {k1: v1}

            # Retest without append
            l = self.cls(path, autosync=True)
            assert len(l) == 1
            assert l[k1] == v1
            assert list(l) == [k1]
            assert dict(l) == {k1: v1}

            # Modify it
            l = self.cls(path, autosync=True)
            l[k1] = v2
            assert dict(l) == {k1: v2}

            l = self.cls(path, autosync=True)
            assert dict(l) == {k1: v2}
            assert l[k1] == v2

    def test__del(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]
        print()
        print("entering test__del")

        with temp_pathname() as path:
            print("path", path, os.path.exists(path))
            d = self.cls(path, {k1: v1, k2: v2, k3: v3}, autosync=True)
            print("temp_path: after init path", path, os.path.exists(path))
            assert dict(d) == {k1: v1, k2: v2, k3: v3}

            del d[k1]
            assert dict(d) == {k2: v2, k3: v3}
            print("test__del:temp_path: containt path", open(path).read())
            d = self.cls(path, autosync=True)
            assert dict(d) == {k2: v2, k3: v3}

    def test__getitem(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]

        with temp_pathname() as path:
            l = self.cls(path, {k1: v1, k2: v2, k3: v3}, autosync=True)
            assert dict(l) == {k1: v1, k2: v2, k3: v3}

            assert l[k1] == v1
            assert l[k3] == v3

            assert dict(l) == {k1: v1, k2: v2, k3: v3}
            assert list(l) == [k1, k2, k3]

    def test_autosync(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            l[k1] = v1

            l = self.cls(dbpath, autosync=True)
            assert l[k1] == v1

            # delete
            del l[k1]
            l = self.cls(dbpath, autosync=True)
            assert len(l) == 0

    def test_contains(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            assert k1 not in l
            assert k2 not in l

            l[k1] = v1
            assert k1 in l
            l = self.cls(dbpath, autosync=True)
            assert k1 in l

            l[k2] = v2
            assert k1 in l
            assert k2 in l
            assert k3 not in l
            l = self.cls(dbpath, autosync=True)
            assert k1 in l
            assert k2 in l
            assert k3 not in l

            l[k3] = v3
            assert k1 in l
            assert k2 in l
            assert k3 in l
            l = self.cls(dbpath, autosync=True)
            assert k1 in l
            assert k2 in l
            assert k3 in l

            del l[k1]
            assert k1 not in l
            l = self.cls(dbpath, autosync=True)
            assert k1 not in l

    def test_file_is_created(self):
        """Assert that the file is created in the constructor"""
        with temp_pathname() as dbpath:
            self.cls(dbpath, autosync=True)
            assert os.path.exists(dbpath)

    def test_initiated_with_existing_dict(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, {k1: v1, k2: v2, k3: v3}, autosync=True)
            assert dict(l) == {k1: v1, k2: v2, k3: v3}
            l = self.cls(dbpath, autosync=True)
            assert dict(l) == {k1: v1, k2: v2, k3: v3}

            del l[k1]
            assert dict(l) == {k2: v2, k3: v3}

            l = self.cls(dbpath, autosync=True)
            assert dict(l) == {k2: v2, k3: v3}

    def test_conflict_args_obj_and_existing_db(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]

        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)  # the file dbpath is created
            with pytest.raises(TypeError):
                # once the file is created we can't initiate it with an existing list
                l = {k1: v1, k2: v2, k3: v3}
                self.cls(dbpath, l, autosync=True)

    def test_double_sync(self):
        """Test that double sync is possible"""
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            l[k1] = v1
            l.sync()
            l.sync()

    def test_nb_sync(self):
        v1, v2, v3 = self.values[:3]
        k1, k2, k3 = self.keys[:3]
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True)
            nb_sync = l._nb_sync
            l[k1] = v1  # sync called here
            l.sync()  # this sync is useless because already synced
            assert l._nb_sync == nb_sync + 1  # check that the previous sync did nothing
            l.update({k2: v2, k3: v3})
            assert (
                l._nb_sync == nb_sync + 2
            )  # check that extend call sync once and not 3 time

    def notttest_unique(self):
        v1, v2, v3 = self.values[:3]

        # test append
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=True)
            l[k1] = v1

            with pytest.raises(ValueError):
                l[k2] = v1

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l.append("a")

        # test insert
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            with pytest.raises(ValueError):
                l.insert(0, "a")

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l.insert(0, "a")

        # test extend
        with temp_pathname() as dbpath:
            # FIXME: if error occure on extend we don't validate first elem?
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            with pytest.raises(ValueError):
                l.extend(["b", "a"])

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l.extend(["b", "a"])

        with temp_pathname() as dbpath:
            # FIXME: if error occure on extend we don't validate first elem?
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            with pytest.raises(ValueError):
                l.extend(["b", "b"])

            l = self.cls(dbpath, autosync=True, unique=True)

        # test setitem
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=True)

            l.append("a")
            l.append("b")
            with pytest.raises(ValueError):
                l[1] = "a"

            l = self.cls(dbpath, autosync=True, unique=True)

            with pytest.raises(ValueError):
                l[1] = "a"

            assert list(l) == ["a", "b"]
            # it is allowed to modify the same value
            l[0] = "a"

    def nottest_unique_as_function(self):
        v1, v2, v3 = self.values[:3]

        # test append
        with temp_pathname() as dbpath:
            l = self.cls(dbpath, autosync=True, unique=lambda x: x.split("=")[0])

            l.append("a=1")
            l.append("b=2")
            l.append("c=3")
            assert list(l) == ["a=1", "b=2", "c=3"]
            l = self.cls(dbpath, autosync=True, unique=lambda x: x.split("=")[0])
            assert list(l) == ["a=1", "b=2", "c=3"]

            with pytest.raises(ValueError):
                l.append("a=5")



import argparse
import inspect
import sys
import typing
import warnings

import argcomplete
import coloring

"""
Group structure:
    name(str)
    title(str)
    description(str)
    commands(dict)
        description(str)
"""


class SubcommandParser:
    def __init__(
        self,
        prog=None,
        *,
        parent=None,
        parser=None,
        autocomplete: bool = None,
        # Help
        colors=None,
        prolog=None,
        description=None,
        epilog=None,
    ):
        """

        Args:
            prog: name of the program
            parent: parent parser
            parser: argparse.ArgParser to use
            description: description of the program
            autocomplete: E,able autocomplete
            colors:

        Advantages over argparse:
            - use add_subcommand instead of using add_parsers then add_subparser
            - run command that will run directly the program
            - better help with groups/colors
            - ease of use autocomplete
        """

        if prog is None:
            prog = sys.argv[0]

        if colors is None:
            colors = False

        if autocomplete is None:
            autocomplete = False

        self.parent = parent
        self.name = prog or sys.argv[0]
        self.subcommands = {}
        self.function = None
        self._argparse_subparsers = None

        # Help
        self.colors = colors
        self.groups = {}
        self.add_group("_ungrouped", title="Other commands")
        self.description = description
        self.prolog = prolog
        self.epilog = epilog
        self.hide = False
        self.autocomplete = autocomplete
        self._group_namespace = (
            set()
        )  # track groupnamespace for subcommands and help_subcommands

        if self.parent is None:
            self.fullname = self.name
        else:
            self.fullname = f"{self.parent.fullname}.{self.name}"

        if parser is None:
            parser = argparse.ArgumentParser(
                prog=prog, description=description, epilog=epilog
            )

        self._argparse_parser = parser
        self._subcommand_depth = 1

        self._print_help = self._argparse_parser.print_help

    def add_argument(self, *args, **kwargs):
        self._argparse_parser.add_argument(*args, **kwargs)

    def add_group(self, name, *, title: str = None, description: str = None):
        if name in self.groups:
            raise KeyError(f"Group {name} already exists")

        if title is None:
            title = name

        if description is None:
            description = ""

        self.groups[name] = {"title": title, "description": description, "commands": {}}

    def add_subcommand(self, command_name, description=None, group=None, hide=False):
        if command_name in self.subcommands:
            raise KeyError(f"command {command_name!r} already exists")

        # TODO: check that help_command is not existing in the same grp
        #  refactor this! we should check all namespace not only the one for grp
        # TODO: test me
        if group and hide:
            raise ValueError(f"Canno't have both group and hide is True")
        if not hide:
            if command_name in self._group_namespace:
                raise KeyError(
                    f"command {command_name!r} already exists as help_subcommand"
                )

            if group is None:
                group = "_ungrouped"

            self.groups[group]["commands"][command_name] = {"description": description}
            self._group_namespace.add(command_name)

        if self._argparse_subparsers is None:
            self._argparse_subparsers = self._argparse_parser.add_subparsers(
                dest=self._get_subcommand_dest_name()
            )

        subcommand_parser = self._argparse_subparsers.add_parser(command_name)

        subcommand_command = SubcommandParser(
            command_name,
            parser=subcommand_parser,
            parent=self,
            description=description,
        )
        subcommand_command._subcommand_depth += self._subcommand_depth
        subcommand_command.hide = hide
        self.subcommands[command_name] = subcommand_command
        return subcommand_command

    def add_help_subcommand(self, command_name, description=None, group=None):
        """Only add this command in the help

        Explanation:
            this could be usefull if you have a lot of commands that are hidden
            and you want to add one help description to group all these commands
        """
        if group is None:
            group = "_ungrouped"

        if command_name in self.subcommands:
            raise KeyError(f"command {command_name!r} already exists")

        if command_name in self._group_namespace:
            raise KeyError(
                f"command {command_name!r} already exists in help_subcommands"
            )

        self.groups[group]["commands"][command_name] = {"description": description}
        self._group_namespace.add(command_name)

    def __getitem__(self, item: str):
        return self.subcommands[item]

    def __str__(self):
        return f"SubcommandParser(<{self.fullname}>)"

    def __repr__(self):
        return self.__str__()

    def print_help(self):
        self._init_help()
        self._print_help()

    def _get_subcommand_dest_name(self, depth: int = None):
        if depth is None:
            depth = self._subcommand_depth
        if depth == 1:
            return "subcommand"
        else:
            return f"subcommand_{depth}"

    def _get_subcommand_name(self, parsed_args, depth: int = None):
        argparse_cmd_name = self._get_subcommand_dest_name(depth)
        return getattr(parsed_args, argparse_cmd_name)

    def parse_args(self, args=None, namespace=None):
        self.check_errors()
        self._init_help()

        if self.autocomplete:
            argcomplete.autocomplete(self._argparse_parser)

        return self._argparse_parser.parse_args(args, namespace=namespace)

    def run(self, args=None):
        """Run the main program"""

        # Parse arguments
        parsed_args = self.parse_args(args)
        # TODO: hook main: self._run_main(parsed_args)  # hook to _run_main

        # Check if there is any subcommand
        if not self.subcommands:
            self.print_help()
            sys.exit(1)

        # get called subcommand
        depth = 1
        subcommand = self
        while True:
            try:

                cmd_name = self._get_subcommand_name(parsed_args, depth=depth)

                if cmd_name is None:
                    self.print_help()
                    sys.exit(1)
                subcommand = subcommand[cmd_name]
                depth += 1
            except AttributeError:
                break

        # If function is None, automatically it (doesn't have subparsers
        #  because we already checked errors on parse_args
        if subcommand.function is None:
            self.print_help()
            sys.exit(0)

        subcommand.function(parsed_args)

    def _init_help(self):
        """If groups exist change the current help"""
        """
        - prolog
        - usage
        - prog description
        - list subcommands
        - epilog
        """
        SPACES = "    "
        LINE_SEP = "\n\n"

        # FIXME: test/me
        # Handle groups
        # =============
        # Groups are already handled in add_subcommand
        if len(self.groups) == 1:
            # If len groups == 1, title = "Commands" else title = "Other commands"
            self.groups["_ungrouped"]["title"] = "Commands"

        # make ungrouped in last position
        _ungrouped = self.groups.pop("_ungrouped")
        self.groups["_ungrouped"] = _ungrouped
        # get the longest command name for pretty print with tabulations
        # flatten all commands names in self.groups()
        commands_names = [
            cmd for group in self.groups.values() for cmd in group["commands"]
        ]

        if len(commands_names):
            max_command_length = len(max(commands_names, key=len))
        else:
            max_command_length = 0

        # header prog name and description
        prog = self.name
        help = ""

        # add prolog
        if self.prolog:
            help += self.prolog + LINE_SEP

        # add usage
        help += f"Usage: {prog} [-h] <subcommand>{LINE_SEP}"

        # add description
        if self.description:
            help += f"{self.description}{LINE_SEP}"

        for group_name, group in self.groups.items():
            if not group["commands"]:
                continue
            group_title = group["title"] + ":"
            if self.colors:
                group_title = coloring.colorize(group_title, c="slate_blue3", s="b")

            group_description = group.get("description", "")

            help += group_title
            if group_description:
                help += " " + group_description
            help += LINE_SEP
            for cmd_name, command_info in group["commands"].items():
                description_command = command_info.get("description", "") or ""
                cmd_txt = cmd_name.ljust(max_command_length)
                if self.colors:
                    cmd_txt = coloring.colorize(cmd_txt, c="medium_purple")
                help += f"{SPACES}{cmd_txt}{SPACES}{description_command}\n"
            help += "\n"

        help += self.epilog or ""
        # change the help function
        self._argparse_parser.print_help = lambda file=None: print(help, file=file)
        # self.main_parser.print_usage = self.main_parser.print_help
        # print(help)
        self._print_help = self._argparse_parser.print_help

    def _print_and_exist(self, msg, status=1):
        print(msg)
        sys.exit(status)

    def iter_allcommands(self):
        """Iter all commands, self included"""
        yield self
        for parser in self.subcommands.values():
            yield from parser.iter_allcommands()

    def check_errors(self):
        """Check if subcommands are correctly built
        This method is called before parse_args/run
        """
        for command in self.iter_allcommands():
            # If function exists it must be callable
            if command.function is not None and not callable(command.function):
                raise TypeError(f"{command.fullname}.function must be callable")

            # Todo: check that function has only one argument

            # If the command don't have any subcommand, it must have a function
            if not command.subcommands and command.function is None:
                raise ValueError(
                    f"Subcommand {command} don't have linked function or should have subpcommands"
                )


class ArgparseSubcmdHelper:
    """argparse class Helper for making subcommands easily
    To add a subcommand x implement the following methods/attributes
        def parser_x(self, parser): add arguments to the subcommand parser
        def run_x(self, args):  run the subcommand
        description_x: description of the subcommand


    cls attributes
        prog:
        _parser_main: add to the main parser
        _run_main(self, args): hook after parsing args, handling main variable before dispatch to subcmd

    Custom help
        groups

    Nested subcommand
        command_x: of Type[ArgparseSubcmdHelper]


    """

    prog = None
    description = None
    groups = None  # FIXME: not yet correctly implemented/tested
    autocomplete = False

    # TODO: add a way to have an optional subcm (dont thtrow and error when called withotu subcmd)
    # TODO: improve the custom help! and test it
    def _parser_main(self, parser):
        """Hook to add arguments to the main parser (general arguments)"""
        pass

    def _run_main(self, args):
        """Hook to execute code before executing the subcommand"""
        pass

    def _init_help(self):
        """If groups exist change the current help"""
        # FIXME: imlement/test me
        if not self.groups:
            return
        SPACES = "    "
        color_help = getattr(self, "color_help", False)
        # header prog name and description
        prog = self.prog or sys.argv[0]
        help = getattr(self, "prolog", "")
        help += f"usage: {prog} [-h] <subcommand>\n\n"

        if self.description:
            help += f"{self.description}\n\n"

        # get the longest command name for pretty print with tabulations
        # flatten all commands names in self.groups()
        grouped_commands_name = [
            cmd for e in self.groups.values() for cmd in e.get("commands", [])
        ]
        all_commands_name = [e[4:] for e in dir(self) if e.startswith("run_")]
        # TODO: add nested commands
        ungrouped_printed_commands = [
            e for e in all_commands_name if e not in grouped_commands_name
        ]
        # max command length for pretty print (aligned)
        max_command_length = len(max(grouped_commands_name, key=len))
        # FIXME: check that ungrouped do not exist
        # add ungrouped commands
        if ungrouped_printed_commands:
            self.groups["ungrouped"] = {
                "commands": ungrouped_printed_commands,
                "title": "Other commands",
                # "description": "commands that do not belong to any group",
            }

        for group_name, group in self.groups.items():
            group_title = group.get("title", group_name) + ":"
            if color_help:
                group_title = coloring.colorize(group_title, c="slate_blue3", s="b")
            commands = group["commands"]
            group_description = group.get("description", "")

            help += group_title
            if group_description:
                help += " " + group_description
            help += "\n\n"
            for cmd in commands:
                # check if the command exist!
                if not hasattr(self, f"run_{cmd}"):
                    raise TypeError(
                        f"Command {cmd!r} can't be in a group because it don't exist (consider implementing 'run_{cmd}')"
                    )
                description_command = getattr(self, f"description_{cmd}", "")
                cmd_txt = cmd.ljust(max_command_length)
                if color_help:
                    cmd_txt = coloring.colorize(cmd_txt, c="medium_purple")
                help += f"{SPACES}{cmd_txt}{SPACES}{description_command}\n"
            help += "\n"

        # change the help function
        self.parser.print_help = lambda file=None: print(help, file=file)
        # self.main_parser.print_usage = self.main_parser.print_help

    def __init__(self, parser=None, subcommand_depth=1):
        # subcommand_depth to allow different level of subcommands and have
        #   the correct dest for subcommands "subcommand" "subcommand_2" "subcommand_3" ...
        self.subcommand_depth = subcommand_depth

        # create main parser
        if parser is None:
            parser = argparse.ArgumentParser(
                prog=self.prog, description=self.description
            )
        self.parser = parser

        # hook to add arguments for the main program
        self._parser_main(self.parser)
        self._build_command()

    def _build_command(self):
        # init the subcommande
        if self.subcommand_depth == 1:
            dest = "subcommand"
        else:
            dest = f"subcommand_{self.subcommand_depth}"

        self.subparsers = self.parser.add_subparsers(dest=dest)
        # Handle all the run_<x> commands
        for subcommand_name in [
            e[len("run_") :] for e in dir(self) if e.startswith("run_")
        ]:
            # Check that the subcommand is callable
            if not callable(getattr(self, "run_" + subcommand_name)):
                raise TypeError(f"Attrbute {'run_'+subcommand_name} must be callable")

            # check that command_<x> and run_<x> dont exist
            if hasattr(self, f"command_{subcommand_name}"):
                raise RuntimeError(
                    f"Both command_{subcommand_name} and run_{subcommand_name} are present"
                )

            # get help if exist
            description = getattr(self, "description_" + subcommand_name, None)
            subparser = self.subparsers.add_parser(
                subcommand_name, help=description, description=description
            )
            parser_name = f"parser_{subcommand_name}"

            # add parser_<cmd> if exist (the parser_<cmd> is not required only run_<cmd> is)
            if hasattr(self, parser_name):
                parser_func = getattr(self, parser_name)
                parser_func(subparser)

        # If parser_x exist but run_x don't raise an exception
        for parser_name in [
            e[len("parser_") :] for e in dir(self) if e.startswith("parser_")
        ]:
            if not hasattr(self, "run_" + parser_name):
                raise RuntimeError(
                    f"parser_{parser_name} method present but run_{parser_name} missing"
                )

        # change help if groups attribute is present
        self._init_help()

        # handle all the command_<x> subcommands
        for subcommand_name in [
            e[len("command_") :] for e in dir(self) if e.startswith("command_")
        ]:
            subcommand_cls: typing.Type[ArgparseSubcmdHelper] = getattr(
                self, "command_" + subcommand_name
            )
            if (not inspect.isclass(subcommand_cls)) or not (
                issubclass(subcommand_cls, ArgparseSubcmdHelper)
            ):
                raise TypeError(
                    f"command_{subcommand_name} attribute must be a class that inherit from {ArgparseSubcmdHelper.__name__!r} not '{type(subcommand_name).__name__}'"
                )

            subcmd_description = getattr(self, f"description_{subcommand_name}", None)
            # TODO: add help_ and description_
            subcmd_parser = self.subparsers.add_parser(
                subcommand_name, description=subcmd_description, help=subcmd_description
            )
            subcommand = subcommand_cls(
                parser=subcmd_parser, subcommand_depth=self.subcommand_depth + 1
            )
            setattr(self, f"_instance_command_{subcommand_name}", subcommand)

    def parse_args(self, args=None):
        if self.autocomplete:
            argcomplete.autocomplete(self.parser)

        return self.parser.parse_args(args)

    def run(self, args=None):
        """Run the main program"""
        if self.autocomplete:
            argcomplete.autocomplete(self.parser)
        parsed_args = self.parser.parse_args(args)
        self._run_main(parsed_args)  # hook to _run_main
        self._run(self, parsed_args)

    def _run(self, command, parsed_args, depth=1):
        """Search the subfunction to run recursivly"""
        if depth == 1:
            subcommand_name_attribute = "subcommand"
        else:
            subcommand_name_attribute = f"subcommand_{depth}"

        subcommand_name = getattr(parsed_args, subcommand_name_attribute)
        if subcommand_name is None:
            command.parser.print_help()
            sys.exit(1)

        if hasattr(command, f"run_{subcommand_name}"):
            subcommand = getattr(command, f"run_{subcommand_name}")
            subcommand(parsed_args)
        elif hasattr(command, f"command_{subcommand_name}"):
            subcommand_instance = getattr(
                command, f"_instance_command_{subcommand_name}"
            )
            self._run(subcommand_instance, parsed_args, depth + 1)

    def main(self, *args, **kwargs):
        """Deprecated run method"""
        warnings.warn(
            "main will be deprecated in future version, use run instead",
            PendingDeprecationWarning,
        )
        self.run(*args, **kwargs)


# TODO: deprecate old argparse
# TODO: add help feature
# TODO: groups with help



import abc
import collections
import json
import os
import shutil
import typing
from typing import Any, List

from koalak.utils import get_prefixed_callables_of_object, temp_pathname

"""
Introduction:
=============
These databases are meant for light usage since they are slow/heavy
and they sinc each time
It is designed to be used as a builtin list or dict


Existing similar project (using files):
=======================================
- sqlite
- text/csv/json
- tinydb
- dataset
- pickledb
- shelve (stdlib)
    - https://code.activestate.com/recipes/576642/ => I used a bit of this
- dbm (key/value)
https://docs.python.org/3/library/persistence.html


Known problems/ To know:
=======================
- Some DB cannot support all types
    - txt: only strings without newlines
    - json: no object
    - ...
- Writeback? (if we use shelve => for the moment not used)
- Autosync
- Attomic sync?


Available databases (implemented or to implement):
==================================================
- list (txt/json/yaml/pickle)
- dict (txt with separator, json, yaml, pickle)
- table? (csv, json, yaml, pickle)
"""

# TODO add Pickle/Yaml on list DB and dict DB
# TODO: Think about delayed opening DB in framework?


class Database(abc.ABC):
    def __init__(self, uri: str, *, mode: str = None, autosync: bool = None):
        """
        Args:
            uri(str): The uri (path) of the DB
            mode(str):
            autosync(bool): if True write to disk after each modification
        """
        if autosync is None:
            autosync = True

        self.uri = uri
        self.autosync = autosync
        self.mode = mode
        self._nb_sync = 0  # TODO: implement me

    def sync(self):
        """Write to disk"""
        if self._synced:
            return
        with temp_pathname() as tmp_dbfile:
            self.dump(tmp_dbfile)

            shutil.move(tmp_dbfile, self.uri)  # atomic commit
            if self.mode is not None:
                os.chmod(self.filename, self.mode)
        self._synced = True
        self._nb_sync += 1

    def _not_synced(self):
        """Set sync at false and autosync"""
        self._synced = False
        if self.autosync:
            self.sync()

    def _check_value(self, e):
        """This method should be implemented on class that have constraint on values"""
        pass

    @abc.abstractmethod
    def dump(self, filepath):
        pass  # pragma: no cover

    @abc.abstractmethod
    def load(self, filepath) -> list:
        pass  # pragma: no cover


class DictDB(collections.UserDict, Database):
    # TODO: add other functions? like count, index ...
    @abc.abstractmethod
    def __init__(self, uri, initdict=None, *, autosync=None, unique=None):
        """Subclass must call super().__init__ and sync()
        We can not call sync on this class because some subclass might need
            extra attributes to call sync() (wich calles dump())

        Notes:
            unique are heavy, at each insertion we will check all values"""
        # TODO: Optimize unique with an inversed dict?

        Database.__init__(self, uri, autosync=autosync)
        # We can not call UserDict.__init__ because it will call self.update()
        #   => self.__setitem__ => self._autosync() => self.sync() => self.dump()

        if unique is None:
            unique = False

        if os.path.exists(uri) and initdict is not None:
            raise TypeError(
                f"DB already created, it can't be initiated with existing dict"
            )

        # We can not call UserDict.__init__ because it will call self.update()
        #   => self.__setitem__ => self._autosync() => self.sync() => self.dump()
        if initdict is None:
            initdict = {}
        self.data = initdict

        self.unique = unique
        self._synced = False  # To optimize useless sync()

        if os.path.exists(uri):
            self.data = self.load(uri)

    def _check_unique(self, key, value):
        if self.unique:
            if self[key] == value:
                return
            for k, v in self.items():
                if v == value:
                    raise ValueError(f"Key")

    def __setitem__(self, key, value):
        # TODO: check value/ check unique
        self._check_value((key, value))
        self.data.__setitem__(key, value)
        self._not_synced()

    def update(self, other):
        # TODO: implement unique
        for k, v in other.items():
            self._check_value((k, v))
            self.data[k] = v
        self._not_synced()

    def __delitem__(self, key):
        self.data.__delitem__(key)
        self._not_synced()

    # TODO: copy


class ListDB(collections.UserList, Database):
    """Abstract class for list database"""

    # 4 Types: Txt, Json, Yaml, Pickle
    # TODO: add constraint feature?
    # TODO: test when the file db exist and an object is present
    # TODO: create generic class WrapperList, wich act like UserList
    #   but will always return a normal list instead of __class__
    #   it will be used for when we always wantto return a normal list like now
    @abc.abstractmethod
    def __init__(self, uri, initlist=None, *, autosync=None, unique=None):
        """Subclass must call super().__init__ and sync()
        We can not call sync on this class because some subclass might need
            extra attributes to call sync() (wich calles dump())"""
        Database.__init__(self, uri, autosync=autosync)
        collections.UserList.__init__(self, initlist)
        if unique is None:
            unique = False

        if os.path.exists(uri) and initlist is not None:
            raise TypeError(
                f"DB already created, it can't be initiated with existing list"
            )

        if initlist is None:
            initlist = []

        self.unique = unique
        self._synced = False  # To optimize useless sync()

        if os.path.exists(uri):
            self.data[:] = self.load(uri)

    def _check_unique(self, e, values=None):
        if values is None:
            values = self.data

        if self.unique:
            if not callable(self.unique):
                if e in values:
                    raise ValueError(
                        f"Unique constraint is present, element already present"
                    )
            else:
                # unique is callable
                unique_e = self.unique(e)
                for e in values:
                    if unique_e == self.unique(e):
                        raise ValueError(
                            f"Unique constraint as function is present, element already present"
                        )

    # FIXME: check __iter__ and __str__

    def __getitem__(self, index):
        # Get item will return a normal list and not DBList!
        return self.data.__getitem__(index)

    def __delitem__(self, key):
        self.data.__delitem__(key)
        self._not_synced()

    def __add__(self, other):
        return self.data + list(other)

    def __radd__(self, other):
        return list(other) + self.data

    def __mul__(self, n):
        return self.data * n

    def __rmul__(self, other):
        return self.__mul__(other)

    # FIXME: __copy__
    def __setitem__(self, index, value):
        # FIXME: try to optimize setitem for the concerned index/check_value/check_unique
        self._check_value(value)

        #  must check other items that the indexed one
        if self.unique:
            all_values_but_current_index = self[:index] + self[index + 1 :]
            self._check_unique(value, values=all_values_but_current_index)

        self.data.__setitem__(index, value)
        self._not_synced()

    def append(self, e: Any):
        self._check_unique(e)
        self._check_value(e)

        self.data.append(e)
        self._not_synced()

    def insert(self, index, value):
        self._check_unique(value)
        self._check_value(value)

        self.data.insert(index, value)
        self._not_synced()

    def pop(self, i=-1):
        return_value = self.data.pop(i)
        self._not_synced()
        return return_value

    def remove(self, value: Any):
        self.data.remove(value)
        self._not_synced()

    def clear(self):
        self.data.clear()
        self._not_synced()

    # TODO: def copy(self)

    def reverse(self):
        self.data.reverse()
        self._not_synced()

    def sort(self, *, key=None, reverse=False):
        self.data.sort(key=key, reverse=reverse)
        self._not_synced()

    def extend(self, iterable):
        for e in iterable:
            self._check_unique(e)
            self._check_value(e)
            self.data.append(e)

        self._not_synced()


class JsonDictDB(DictDB):
    # TODO: add update optimized (like extend)
    def __init__(self, uri, obj=None, *, autosync=None, unique=None, indent=4):
        super().__init__(uri, obj, autosync=autosync, unique=unique)

        self.indent = indent
        self.sync()

    def dump(self, filepath):
        with open(filepath, "w") as fileobj:
            json.dump(self.data, fileobj, indent=self.indent)

    def load(self, filepath) -> list:
        with open(filepath) as fileobj:
            l = json.load(fileobj)
            if not isinstance(l, dict):
                raise TypeError(f"Loaded DB is not a dict but {type(l)}")
            return l


class JsonListDB(ListDB):
    def __init__(self, uri, obj=None, *, autosync=None, unique=None, indent=4):
        super().__init__(uri, obj, autosync=autosync, unique=unique)

        self.indent = indent
        self.sync()

    def dump(self, filepath):
        with open(filepath, "w") as fileobj:
            json.dump(self.data, fileobj, indent=self.indent)

    def load(self, filepath) -> list:
        with open(filepath) as fileobj:
            l = json.load(fileobj)
            if not isinstance(l, list):
                raise TypeError(f"Loaded DB is not a list but {type(l)}")
            return l


class TxtListDB(ListDB):
    def __init__(self, uri, initlist=None, *, autosync=None, unique=None):
        super().__init__(uri, initlist, autosync=autosync, unique=unique)
        self.sync()

    def _check_value(self, e):
        if not isinstance(e, str):
            raise TypeError(f"Added element must be of type str and not {type(e)!r}")
        if "\n" in e:
            raise ValueError(f"Added element must not contain newline")

    # TODO: we can optimize this class append can just write
    #   but it is not that simple with autosync ... see removed code

    def dump(self, filepath):
        with open(filepath, "w") as fileobj:
            for e in self.data:
                fileobj.write(e)
                fileobj.write("\n")

    def load(self, filepath) -> list:
        with open(filepath) as fileobj:
            l = fileobj.read().splitlines()
            return l


class TxtDictDB(DictDB):
    def __init__(self, uri, initlist=None, *, sep=":", autosync=None, unique=None):
        self.sep = sep
        super().__init__(uri, initlist, autosync=autosync, unique=unique)
        self.sync()

    def _check_value(self, e):
        k, v = e
        if not isinstance(v, str):
            raise TypeError(f"Added value must be of type str and not {type(v)!r}")
        if "\n" in v:
            raise ValueError(f"Added value must not contain newline")
        if not isinstance(k, str):
            raise TypeError(f"Added key must be of type str and not {type(k)!r}")
        if "\n" in k:
            raise ValueError(f"Added key must not contain newline")
        if self.sep in k:
            raise ValueError(f"Key can't contain the seprator {self.sep!r}")

    # TODO: we can optimize this class append can just write
    #   but it is not that simple with autosync ... see removed code

    def dump(self, filepath):
        with open(filepath, "w") as fileobj:
            for k, v in self.data.items():
                fileobj.write(k)
                fileobj.write(self.sep)
                fileobj.write(v)
                fileobj.write("\n")

    def load(self, filepath) -> list:
        d = {}
        # TODO: what if a key exist twice?
        with open(filepath) as fileobj:
            for line in fileobj:
                k, v = line[:-1].split(self.sep, 1)
                d[k] = v

        return d


# TODO: test dict unique



from inspect import _ParameterKind
from typing import Dict

import attr
import docstring_parser
from koalak import SubcommandParser
from koalak.decorators import optionalargs


@attr.s
class APIMethod:
    name = attr.ib()
    description = attr.ib()
    method = attr.ib()
    args = attr.ib()
    kwargs = attr.ib()


@attr.s
class APIParameter:
    name = attr.ib()
    description = attr.ib()
    default = attr.ib()
    annotation = attr.ib()


class BudgetManager:
    def __init__(self, money=1000):
        self.transactions = []
        self.money = money

    # API rest:  /api/v1/outcom/
    def outcom(self, cost: int, description: str = ""):
        """
        Args:
            cost(cost): lol
            description(str): haha
        """
        self.transactions.append(cost)
        self.money -= cost

    def show(self):
        for t in self.transactions:
            print(t)


import inspect

s = inspect.signature(BudgetManager.outcom)


class MultiAPI:
    def __init__(self, name):
        self.name = name
        self.methods: Dict[str, APIMethod] = {}
        self.maincli = None

    def build(self):
        self.build_cli()

    @optionalargs
    def add(
        self, method, *, name=None, description=None, http_method=None, help_group=None
    ):
        if name is None:
            name = method.__name__

        if name in self.methods:
            raise ValueError(f"Method {name} already exists")

        parsed_docstring = docstring_parser.parse(method.__doc__)
        if description is None:
            description = parsed_docstring.short_description

        args = []
        kwargs = {}
        signature = inspect.signature(method)
        for i, param in enumerate(signature.parameters.values()):
            if i == 0 and param.name == "self":
                continue
            api_param = APIParameter(
                param.name,
                description="",
                default=param.default,
                annotation=param.annotation,
            )
            if param.kind in [
                _ParameterKind.POSITIONAL_OR_KEYWORD,
                _ParameterKind.POSITIONAL_ONLY,
            ]:
                args.append(api_param)
            else:
                kwargs[param.name] = api_param

        method_api = APIMethod(
            name, method=method, description=description, args=args, kwargs=kwargs
        )
        self.methods[name] = method_api
        print(method_api)
        return method

    def build_cli(self):
        self.maincli = SubcommandParser(self.name)
        for method in self.methods.values():
            subcmd = self.maincli.add_subcommand(
                method.name, description=method.description
            )
            subcmd.function = method.method
            for param in method.args:
                subcmd.add_argument(param.name, type=param.annotation)
            for param in method.kwargs.values():
                subcmd.add_argument(f"--{param.name}", type=param.annotation)


multiapi = MultiAPI("budgetm")


class BudgetManager:
    def __init__(self, money=1000):
        self.transactions = []
        self.money = money

    @multiapi.add(http_method="POST", help_group="basics")
    def outcom(self, cost: int, description: str = ""):
        """
        Add a new outcom to the current profil

        Args:
            cost(int): cost to add
            description(str): description of the transaction
        """
        self.transactions.append(-cost)
        self.money -= cost

    @multiapi.add
    def show(self):
        for t in self.transactions:
            print(t)

    @multiapi.add(http_method="POST", help_group="basics")
    def incom(self, cost: int, description: str = ""):
        """
        Add a new outcom to the current profil

        Args:
            cost(int): cost to add
            description(str): description of the transaction
        """
        self.transactions.append(cost)
        self.money += cost


# FIXME: how to add groups!?
# FIXME: how to chose which instance to use
# FIXME: how to add description of prog
def manual_build():
    budgetm_cmd = SubcommandParser("budgetm", description="TODO")
    outcom_cmd = budgetm_cmd.add_subcommand("outcom")
    outcom_cmd.add_argument("cost", type=int, description="Adding things")

    show_cmd = budgetm_cmd.add_subcommand("show")


if __name__ == "__main__":
    multiapi.build()
    multiapi.maincli.run()



import os
import sys
import tempfile

import koalak
import pytest
from koalak.utils import temp_pathname, tmp_module


def test_api_working():
    koalak.mkpluginmanager()
    koalak.mkpluginmanager("something")


def test_unnamed_without_framework():
    plugins = koalak.mkpluginmanager()

    @plugins.mkbaseplugin
    class BasePlugin:
        pass

    class SimplePlugin(BasePlugin):
        name = "simple"

    assert plugins["simple"] == SimplePlugin
    assert "simple" in plugins
    assert SimplePlugin in plugins
    assert list(plugins) == [SimplePlugin]
    assert len(plugins) == 1


def test_simple_without_framework():
    plugins = koalak.mkpluginmanager("plugin")

    @plugins.mkbaseplugin
    class BasePlugin:
        pass

    class SimplePlugin(BasePlugin):
        name = "simple"

    assert plugins["simple"] == SimplePlugin
    assert "simple" in plugins
    assert SimplePlugin in plugins
    assert list(plugins) == [SimplePlugin]


def test_two_plugins_without_framework():
    plugins = koalak.mkpluginmanager("plugin")

    @plugins.mkbaseplugin
    class BasePlugin:
        pass

    class OnePlugin(BasePlugin):
        name = "one"

    class TwoPlugin(BasePlugin):
        name = "two"

    assert plugins["one"] == OnePlugin
    assert "one" in plugins
    assert OnePlugin in plugins

    assert plugins["two"] == TwoPlugin
    assert "two" in plugins
    assert TwoPlugin in plugins

    assert list(plugins) == [OnePlugin, TwoPlugin]
    assert len(plugins) == 2


def test_with_framework():
    framework = koalak.mkframework()
    plugins = framework.mkpluginmanager("tools")

    @plugins.mkbaseplugin
    class BasePlugin:
        pass

    class SimplePlugin(BasePlugin):
        name = "simple"

    assert plugins["simple"] == SimplePlugin
    assert "simple" in plugins
    assert SimplePlugin in plugins

    assert list(plugins) == [SimplePlugin]
    assert len(plugins) == 1
    assert framework.plugin_managers["tools"] is plugins
    assert len(framework.plugin_managers) == 1


@pytest.mark.skip
def test_error_abstract_method():
    framework = koalak.mkframework()

    @framework.mkpluginmanager("test")
    class BaseTest:
        @framework.abstract
        def x(self):
            pass

    class XTest(BaseTest):
        name = "x"

        def x(self):
            pass

    with pytest.raises(TypeError):

        class YTest(BaseTest):
            name = "y"


def test_abstract_subplugin():
    framework = koalak.mkframework()
    plugins = framework.mkpluginmanager("tools")

    @plugins.mkbaseplugin
    class BasePlugin:
        # This plugin requires the description attribute
        description = plugins.attr(type=str)

    # Plugin A must have the attribute "name"
    with pytest.raises(TypeError):

        class APlugin(BasePlugin):
            name = "A"

    # Abstract class are not checked!
    class AbstractPlugin(BasePlugin):
        abstract = True
        description = "Hey"

    # don't need description since abstract class have it
    class BPlugin(AbstractPlugin):
        name = "B"


def test_load_plugins():
    # TODO: check if this test is correct
    # TODO: Test with existing directory and without
    with tempfile.TemporaryDirectory() as plugin_path:
        plugins = koalak.mkpluginmanager(homepath=plugin_path)

        @plugins.mkbaseplugin
        class BasePlugin:
            pass

        class APlugin:
            name = "A"

        # Temporally create a module so it can be importable by the plugin
        with tmp_module(context={"BasePlugin": BasePlugin}) as module:
            name_module = module.__name__
            # print(name_module, name_module in sys.modules)
            data_file = f"""from {name_module} import BasePlugin
class BPlugin(BasePlugin):
    name = 'B'
    """
            with open(os.path.join(plugin_path, "plugin.py"), "w") as f:
                print("in path", os.path.join(plugin_path, "plugin.py"))
                print("data_file", data_file)
                f.write(data_file)

            # Init folder and plugins
            plugins.init()
            assert plugins["B"].name == "B"  # Just check that "B" is loaded in plugnis


def test_init_without_homepath():
    """We can not initiate a plugin manager without homepath"""
    plugins = koalak.mkpluginmanager()
    with pytest.raises(TypeError):
        plugins.init()  # No error should be raised


def test_double_init():
    """We can init twice"""
    with temp_pathname() as pathname:
        plugins = koalak.mkpluginmanager(homepath=pathname)
        plugins.init()
        with pytest.raises(TypeError):
            plugins.init()


def test_init_with_homepath():
    """Homepath is correctly created"""
    with temp_pathname() as pathname:
        plugins = koalak.mkpluginmanager(homepath=pathname)
        plugins.init()
        assert os.path.isdir(plugins.homepath)


def test_constraint_attribute():
    """Test that attribute "help" is required"""
    plugins = koalak.mkpluginmanager()

    @plugins.mkbaseplugin
    class BasePlugin:
        help = plugins.attr()  # help attribute is required

    class APlugin(BasePlugin):
        name = "A"
        help = "Something"

    class BPlugin(BasePlugin):
        name = "A"
        help = 12  # can be of any type

    with pytest.raises(TypeError):
        # Must define help
        class CPlugin(BasePlugin):
            name = "B"


def test_constraint_attr_type():
    plugins = koalak.mkpluginmanager()

    @plugins.mkbaseplugin
    class BasePlugin:
        help = plugins.attr(type=str)

    class APlugin(BasePlugin):
        name = "A"
        help = "Something"

    with pytest.raises(TypeError):
        # Help must be a string
        class BPlugin(BasePlugin):
            name = "B"
            help = 5


def test_constraint_attr_inheritable_true():
    plugins = koalak.mkpluginmanager()

    @plugins.mkbaseplugin
    class BasePlugin:
        x = plugins.attr()  # By default inheritable is True

    class APlugin(BasePlugin):
        name = "A"
        x = 1

    class BPlugin(APlugin):
        name = "B"


def test_constraint_attr_inheritable_false():
    plugins = koalak.mkpluginmanager()

    @plugins.mkbaseplugin
    class BasePlugin:
        x = plugins.attr(inheritable=False)

    class APlugin(BasePlugin):
        name = "A"
        x = 1

    with pytest.raises(TypeError):

        class BPlugin(APlugin):
            name = "B"


def test_constraint_attr_choices():
    plugins = koalak.mkpluginmanager()

    @plugins.mkbaseplugin
    class BasePlugin:
        help = plugins.attr(type=int, choices=[1, 2])

    class APlugin(BasePlugin):
        name = "A"
        help = 1

    with pytest.raises(ValueError):
        # Help must be 1 or 2
        class BPlugin(BasePlugin):
            name = "B"
            help = 3


def test__repr__and__str__():
    pm = koalak.mkpluginmanager()
    assert repr(pm) == str(pm) == "<PluginManager>"

    pm = koalak.mkpluginmanager("tools")
    assert repr(pm) == str(pm) == "<PluginManager [tools]>"



import inspect

import pytest
from koalak import add_post_init


def addx_post_init(self):
    self.x = True


def addy_post_init(self):
    self.y = True


def addz_post_init(self):
    self.z = True


def test_cls_with_simple_init():
    @add_post_init(addx_post_init)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert a.a is True
    assert a.x is True


def test_cls_with_complex_init():
    @add_post_init(addx_post_init)
    class A:
        def __init__(self, a, b, *, c, d):
            self.a = a
            self.b = b
            self.c = c
            self.d = d

    a = A(1, 2, c=3, d=4)
    assert a.a == 1
    assert a.b == 2
    assert a.c == 3
    assert a.d == 4
    assert a.x is True


def test_multiple_post_inits():
    # When the cls has multiple post_inits
    # execute them all
    @add_post_init(addx_post_init)
    @add_post_init(addy_post_init)
    @add_post_init(addz_post_init)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert a.a is True
    assert a.x is True
    assert a.y is True
    assert a.z is True


def test_inheritence():
    class A:
        def __init__(self):
            self.a = True

    @add_post_init(addx_post_init)
    class B(A):
        def __init__(self):
            super().__init__()
            self.b = True

    a = A()
    assert a.a is True
    assert not hasattr(a, "b")
    assert not hasattr(a, "x")

    b = B()
    assert b.a is True
    assert b.b is True
    assert b.x is True


def test_cls_without_init():
    @add_post_init(addx_post_init)
    class A:
        pass

    a = A()
    assert a.x is True


def test_cls_without_init_and_parent_init():
    # the add_post_init must not modify the parent init
    class A:
        def __init__(self):
            self.a = True

    @add_post_init(addx_post_init)
    class B(A):
        pass

    a = A()
    assert a.a is True
    assert not hasattr(a, "x")

    b = B()
    assert b.a is True
    assert b.x is True


def test_if_signature_changed():
    class A:
        def __init__(self, a: str, b: int, *, c: bool, d: str):
            """my init"""
            self.a = a
            self.b = b
            self.c = c
            self.d = d

    original_signature = inspect.signature(A)
    original_doc = A.__init__.__doc__
    # apply decorator dynamically
    A = add_post_init(addx_post_init)(A)

    assert inspect.signature(A) == original_signature
    assert A.__init__.__doc__ == original_doc

    a = A("a", 1, c=True, d="b")
    assert a.a == "a"
    assert a.b == 1
    assert a.c is True
    assert a.d == "b"
    assert a.x is True


def test_signature_of_created_init():
    class A:
        pass

    original_signature = inspect.signature(A)
    # apply decorator dynamically
    A = add_post_init(addx_post_init)(A)
    a = A()
    assert a.x is True

    with pytest.raises(TypeError):
        # Verify that signature is not *args, **kwargs
        A(1)

    assert inspect.signature(A) == original_signature



import koalak
import pytest
from koalak import exceptions, frameworks


def test_error_framework_already_exist():
    framework_name = koalak.generate_unique_framework_name()
    framework1 = koalak.mkframework(framework_name)  # noqa: F841

    with pytest.raises(exceptions.FrameworkAlreadyExistException):
        framework2 = koalak.mkframework(framework_name)  # noqa: F841


def test_contains_frameworks():
    framework1 = koalak.mkframework(koalak.generate_unique_framework_name())
    framework2 = koalak.mkframework(koalak.generate_unique_framework_name())

    assert framework1 in koalak.frameworks
    assert framework1.name in koalak.frameworks
    assert framework2 in koalak.frameworks
    assert framework2.name in koalak.frameworks


def test_get_frameworks():
    framework1 = koalak.mkframework(koalak.generate_unique_framework_name())
    framework2 = koalak.mkframework(koalak.generate_unique_framework_name())

    assert framework1 is koalak.frameworks[framework1.name]
    assert framework2 is koalak.frameworks[framework2.name]
    assert framework1 in list(koalak.frameworks)
    assert framework2 in list(koalak.frameworks)



import os
import tempfile

import pytest
from koalak import D, F, mkframework
from koalak.bases import normalize_home_structure

# TODO: unskip test and continue implementations


@pytest.mark.skip
def test_init_home_dict_files():
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = {
            "a": "",
            "b": "",
            "c": ""
        }
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)
        for path in ["a", "b", "c"]:
            path = os.path.join(home_path, path)
            assert os.path.isfile(path)
            with open(path) as f:
                assert f.read() == ""


@pytest.mark.skip
def test_init_home_dict_dirs():
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = {
            "a": {},
            "b": {},
            "c": {}
        }
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)
        for path in ["a", "b", "c"]:
            path = os.path.join(home_path, path)
            assert os.path.isdir(path)


@pytest.mark.skip
def test_init_home_dict_files_and_dirs():
    # str is a directory
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = {
            "a": "",
            "b": "",
            "c": {},
            "d": {},
            "e": {},
            "f": {}

        }
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)

        for path in ["a", "b"]:
            path = os.path.join(home_path, path)
            assert os.path.isfile(path)
            with open(path) as f:
                assert f.read() == ""

        for path in ["c", "d", "e", "f"]:
            path = os.path.join(home_path, path)
            assert os.path.isdir(path)


@pytest.mark.skip
def test_init_home_recurse():
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = {
            "a": "",
            "b": "",
            "c": {},
            "d": {
                "e": {},
                "f": ""
            },
            "g": {
                "h": {
                    "i": {
                        "j": {
                            "k": ""
                        }
                    }
                }
            }
        }
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)

        assert os.path.isfile(os.path.join(home_path, "a"))
        assert os.path.isfile(os.path.join(home_path, "b"))
        assert os.path.isdir(os.path.join(home_path, "c"))
        assert os.path.isdir(os.path.join(home_path, "d"))
        assert os.path.isdir(os.path.join(home_path, "d", "e"))
        assert os.path.isfile(os.path.join(home_path, "d", "f"))
        assert os.path.isdir(os.path.join(home_path, "g"))
        assert os.path.isdir(os.path.join(home_path, "g", "h"))
        assert os.path.isdir(os.path.join(home_path, "g", "h", "i"))
        assert os.path.isdir(os.path.join(home_path, "g", "h", "i", "j"))
        assert os.path.isfile(os.path.join(home_path, "g", "h", "i", "j", "k"))


@pytest.mark.skip
def test_init_home_dict_real_example():
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = {
            ".bashrc": "",
            "bin": {},
            "github": {
                "jwtattack": {},
                "gobuster": {},
                "nmap": {}
            },
            "tools": {
                "seclist": {}
            }

        }
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)

        assert os.path.isfile(os.path.join(home_path, ".bashrc"))
        assert os.path.isdir(os.path.join(home_path, "github"))
        assert os.path.isdir(os.path.join(home_path, "github", "jwtattack"))
        assert os.path.isdir(os.path.join(home_path, "github", "gobuster"))
        assert os.path.isdir(os.path.join(home_path, "github", "nmap"))
        assert os.path.isdir(os.path.join(home_path, "tools"))
        assert os.path.isdir(os.path.join(home_path, "tools", "seclist"))


def test_init_home_F_D_real_example():
    """Test real example with F/D syntax"""
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = [
            F(".bashrc"),
            D("bin"),
            D(
                "github",
                [D("jwtattack"), D("gobuster"), D("nmap"), F("note.txt")]
            ),
            D("tools", [D("seclist")])
        ]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)

        assert os.path.isfile(os.path.join(home_path, ".bashrc"))
        assert os.path.isdir(os.path.join(home_path, "bin"))
        assert os.path.isdir(os.path.join(home_path, "github"))
        assert os.path.isdir(os.path.join(home_path, "github", "jwtattack"))
        assert os.path.isdir(os.path.join(home_path, "github", "gobuster"))
        assert os.path.isdir(os.path.join(home_path, "github", "nmap"))
        assert os.path.isfile(os.path.join(home_path, "github", "note.txt"))
        assert os.path.isdir(os.path.join(home_path, "tools"))
        assert os.path.isdir(os.path.join(home_path, "tools", "seclist"))


@pytest.mark.skip
def test_init_home_list_files():
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = ["a", "b", "c"]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)
        for path in ["a", "b", "c"]:
            path = os.path.join(home_path, path)
            assert os.path.isfile(path)
            with open(path) as f:
                assert f.read() == ""


def test_init_home_F_content():
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = [F("a"), F("b", content="Hello")]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)
        assert os.path.isfile(os.path.join(home_path, "a"))
        assert open(os.path.join(home_path, "a")).read() == ""
        assert os.path.isfile(os.path.join(home_path, "b"))
        assert open(os.path.join(home_path, "b")).read() == "Hello"


def test_init_home_F_substitute():
    """Test variables substitutions"""
    with tempfile.TemporaryDirectory() as dirname:
        # fmt: off
        home = [F("a"), F("b", content="my age is $age", substitute=True)]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(
            homepath=home_path, home_structure=home, variables={"age": 30}
        )
        framework.init()

        assert os.path.isdir(home_path)
        assert os.path.isfile(os.path.join(home_path, "a"))
        assert open(os.path.join(home_path, "a")).read() == ""
        assert os.path.isfile(os.path.join(home_path, "b"))
        assert open(os.path.join(home_path, "b")).read() == "my age is 30"


def test_init_home_F_src():
    """Srs attribute take the content of a file from an existing file"""
    with tempfile.TemporaryDirectory() as dirname, tempfile.NamedTemporaryFile(
        "w"
    ) as tmpfile:
        # write on file
        tmpfile.file.write("World!")
        tmpfile.file.flush()
        # fmt: off
        home = [F("a"), F("b", content="Hello"), F("c", src=tmpfile.name)]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        framework = mkframework(homepath=home_path, home_structure=home)
        framework.init()

        assert os.path.isdir(home_path)
        assert os.path.isfile(os.path.join(home_path, "a"))
        assert open(os.path.join(home_path, "a")).read() == ""
        assert os.path.isfile(os.path.join(home_path, "b"))
        assert open(os.path.join(home_path, "b")).read() == "Hello"
        assert os.path.isfile(os.path.join(home_path, "c"))
        assert open(os.path.join(home_path, "c")).read() == "World!"


# ========================== #
# test _normalize_home_nodes #
# ========================== #


def test__normalize_home_nodes_list_of_str():
    nodes = ["a", "b", "c"]
    expected = [F("a"), F("b"), F("c")]
    result = normalize_home_structure(nodes)
    assert expected == result


def test__normalize_home_nodes_list_of_str_and_F_D():
    nodes = ["a", "b", "c", F("d"), D("e")]
    expected = [F("a"), F("b"), F("c"), F("d"), D("e")]
    result = normalize_home_structure(nodes)
    assert expected == result


def test__normalize_home_nodes_list_of_str_with_dirs():
    nodes = ["a", "b", "c/"]
    expected = [F("a"), F("b"), D("c")]
    result = normalize_home_structure(nodes)
    assert expected == result


@pytest.mark.skip
def test__normalize_home_nodes_dict():
    # fmt: off
    nodes = {
        "a": {},
        "b": {},
        "c": "",
    }
    # fmt: on
    expected = [D("a"), D("b"), F("c")]
    result = normalize_home_structure(nodes)
    assert expected == result


"""
x = ["a", "b", "c"] +  [F("a"), F("b"), F("C")]
home = {
            ".bashrc": "",
            "bin": {},
            "github": {
                "jwtattack": {},
                "gobuster": {},
                "nmap": {}
            },
            "tools": {
                "seclist": {}
            }

        }
home = [
    F(".bashrc"),
    D("bin"),
    D("github", D("jwtattack"))
]
"""
